from pathlib import Path
from unittest import mock

import dash

from webviz_config import WebvizConfigTheme
from webviz_config.common_cache import CACHE
from webviz_config.themes import default_theme
from webviz_config.generic_plugins import _syntax_highlighter


def test_create_themed_layout():
    # Make an instance of the theme class
    test_theme = WebvizConfigTheme(theme_name="test")
    # Add a plotly theme to the new theme
    test_theme.plotly_theme = {
        "layout": {
            "font": {"family": "Arial"},
            "colorscale": {
                "diverging": [
                    [0, "rgb(255, 0, 0)"],
                    [0.5, "rgb(255, 255, 0)"],
                    [1, "rgb(255, 255, 255)"],
                ],
                "sequential": [
                    [0, "rgb(0, 0, 255)"],
                    [0.5, "rgb(0, 255, 0)"],
                    [1, "rgb(255, 0, 0)"],
                ],
            },
            "xaxis": {"title": {"text": "Hi", "font": {"size": 10}}},
            "yaxis2": {"color": "#000"},
        },
    }
    # Define a layout that should be themed
    test_layout = {
        "font": {"size": 12},
        "colorscale": {
            "sequential": [
                [0, "rgb(0, 0, 0)"],
                [0.25, "rgb(255, 0, 0)"],
                [0.5, "rgb(0, 255, 0)"],
                [0.75, "rgb(0, 0, 255)"],
                [1, "rgb(255, 255, 255)"],
            ],
        },
        "xaxis": {"title": {"text": "New", "standoff": 2}, "type": "log"},
        "xaxis2": {},
        "yaxis": {"color": "#111"},
    }
    # The expected layout after theming is applied
    expected_themed_layout = {
        "font": {"family": "Arial", "size": 12},
        "colorscale": {
            "diverging": [
                [0, "rgb(255, 0, 0)"],
                [0.5, "rgb(255, 255, 0)"],
                [1, "rgb(255, 255, 255)"],
            ],
            "sequential": [
                [0, "rgb(0, 0, 0)"],
                [0.25, "rgb(255, 0, 0)"],
                [0.5, "rgb(0, 255, 0)"],
                [0.75, "rgb(0, 0, 255)"],
                [1, "rgb(255, 255, 255)"],
            ],
        },
        "xaxis": {
            "title": {"text": "New", "font": {"size": 10}, "standoff": 2},
            "type": "log",
        },
        "xaxis2": {"title": {"text": "Hi", "font": {"size": 10}}},
        "yaxis": {"color": "#111"},
        "yaxis2": {"color": "#000"},
    }

    # Make a themed layout using the create_themed_layout method
    themed_layout = test_theme.create_themed_layout(test_layout)

    # Verify that the themed layout is as expected
    assert themed_layout == expected_themed_layout

    # Verify that the theme layout is unchanged (same as when set)
    assert test_theme.plotly_theme == {
        "layout": {
            "font": {"family": "Arial"},
            "colorscale": {
                "diverging": [
                    [0, "rgb(255, 0, 0)"],
                    [0.5, "rgb(255, 255, 0)"],
                    [1, "rgb(255, 255, 255)"],
                ],
                "sequential": [
                    [0, "rgb(0, 0, 255)"],
                    [0.5, "rgb(0, 255, 0)"],
                    [1, "rgb(255, 0, 0)"],
                ],
            },
            "yaxis2": {"color": "#000"},
            "xaxis": {"title": {"text": "Hi", "font": {"size": 10}}},
        },
    }

    # Verify that the test_layout unchanged (same as when set)
    assert test_layout == {
        "font": {"size": 12},
        "colorscale": {
            "sequential": [
                [0, "rgb(0, 0, 0)"],
                [0.25, "rgb(255, 0, 0)"],
                [0.5, "rgb(0, 255, 0)"],
                [0.75, "rgb(0, 0, 255)"],
                [1, "rgb(255, 255, 255)"],
            ],
        },
        "xaxis": {"title": {"text": "New", "standoff": 2}, "type": "log"},
        "yaxis": {"color": "#111"},
        "xaxis2": {},
    }


def test_create_themed_layout_on_app(dash_duo):
    # Test themed with a running app
    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    code_file = Path("./tests/data/basic_example.yaml")
    with mock.patch(
        "webviz_config.generic_plugins._syntax_highlighter.get_path"
    ) as mock_path:
        mock_path.return_value = code_file
        page = _syntax_highlighter.SyntaxHighlighter(app, code_file)
        app.layout = page.layout
        dash_duo.start_server(app)
        # empty input dict should return the theme layout
        assert default_theme.plotly_theme[
            "layout"
        ] == default_theme.create_themed_layout({})
        # empty input dict should return the theme layout
        specified_layout = {
            "colorway": ["#FFFFFF", "#0000FF"],
            "xaxis": {"title": {"text": "Title"}, "type": "log", "constrain": "domain"},
            "legend": {"font": {"family": "Arial"}},
        }
        # create a new layout from specified and theme layouts
        new_layout = default_theme.create_themed_layout(specified_layout)
        # test some values
        assert new_layout["colorway"] == specified_layout["colorway"]
        assert new_layout["xaxis"]["title"]["text"] == "Title"
        assert not dash_duo.get_logs(), "browser console should contain no error"



import base64
from pathlib import Path
from typing import List, Optional

import pandas as pd
from dash import Dash
import dash_pivottable

from .. import WebvizPluginABC, EncodedFile
from ..webviz_store import webvizstore
from ..common_cache import CACHE


class PivotTable(WebvizPluginABC):
    """Adds a pivot table to the webviz instance, using tabular data from a \
        provided csv file.

---

* **`csv_file`:** Path to the csv file containing the tabular data. Either absolute \
                  path or relative to the configuration file.
* **`options`:** Additional options for the plot. See [dash-pivottable documentation]\
    (https://github.com/plotly/dash-pivottable#references) for all possible options.
"""

    def __init__(self, app: Dash, csv_file: Path, options: dict = None):

        super().__init__()

        self.csv_file = csv_file
        self.options = options if options is not None else {}

        self.set_callbacks(app)

    def add_webvizstore(self) -> List[tuple]:
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def layout(self) -> dash_pivottable.PivotTable:
        return generate_table(get_data(self.csv_file), **self.options)

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(self.plugin_data_output, self.plugin_data_requested)
        def _user_download_data(data_requested: Optional[int]) -> Optional[EncodedFile]:
            return (
                {
                    "filename": "pivot-table.csv",
                    "content": base64.b64encode(
                        get_data(self.csv_file).to_csv(index=False).encode()
                    ).decode("ascii"),
                    "mime_type": "text/csv",
                }
                if data_requested
                else None
            )


def generate_table(dframe: pd.DataFrame, **options: str) -> dash_pivottable.PivotTable:
    return dash_pivottable.PivotTable(
        data=[dframe.columns.to_list()] + dframe.values.tolist(), **options
    )


@CACHE.memoize()
@webvizstore
def get_data(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file)



import base64
import inspect
from pathlib import Path
from collections import OrderedDict
from typing import Optional, List, Dict, Any

import numpy as np
import pandas as pd
from plotly.graph_objects import Figure
import plotly.express as px
from dash import html, dcc, Input, Output, Dash
import webviz_core_components as wcc

from .. import WebvizPluginABC, WebvizSettings, EncodedFile
from ..webviz_store import webvizstore
from ..common_cache import CACHE


# pylint: disable=too-many-arguments
class TablePlotter(WebvizPluginABC):
    """Adds a plotter to the webviz instance, using tabular data from a provided csv file.
If feature is requested, the data could also come from a database.

---

* **`csv_file`:** Path to the csv file containing the tabular data. \
                  Either absolute path or relative to the configuration file.
* **`plot_options`:** A dictionary of plot options to initialize the plot with.
* **`filter_cols`:** Dataframe columns that can be used to filter data.
* **`filter_defaults`:** A dictionary with column names as keys, \
                         and a list of column values that should be preselected in the filter. \
                         If a columm is not defined, all values are preselected for the column.
* **`column_color_discrete_maps`:** A dictionary with column names as keys, \
                                    each key containing a new dictionary with the columns \
                                    unique values as keys, and the color they should be \
                                    plotted with as value. Hex values needs quotes '' \
                                    to not be read as a comment.
* **`lock`:** If `True`, only the plot is shown, \
              all dropdowns for changing plot options are hidden.
"""

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csv_file: Path,
        plot_options: dict = None,
        filter_cols: list = None,
        filter_defaults: dict = None,
        column_color_discrete_maps: dict = None,
        lock: bool = False,
    ) -> None:

        super().__init__()

        self.plot_options = plot_options if plot_options else {}
        self.lock = lock
        self.csv_file = csv_file
        self.data = get_data(self.csv_file)
        self.set_filters(filter_cols)
        self.columns = list(self.data.columns)
        self.numeric_columns = list(
            self.data.select_dtypes(include=[np.number]).columns
        )
        self.filter_defaults = filter_defaults
        self.column_color_discrete_maps = column_color_discrete_maps
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.set_callbacks(app)

    def set_filters(self, filter_cols: Optional[list]) -> None:
        self.filter_cols = []
        self.use_filter = False
        if filter_cols:
            for col in filter_cols:
                if col in self.data.columns:
                    if self.data[col].nunique() != 1:
                        self.filter_cols.append(col)
            if self.filter_cols:
                self.use_filter = True

    def add_webvizstore(self) -> List[tuple]:
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def plots(self) -> dict:
        """A dict of available plots and their options"""
        return {
            "scatter": ["x", "y", "size", "color", "facet_col"],
            "histogram": [
                "x",
                "color",
                "facet_col",
                "barmode",
                "barnorm",
                "histnorm",
            ],
            "bar": ["x", "y", "color", "facet_col", "barmode"],
            "pie": ["values", "names"],
            "scatter_3d": ["x", "y", "z", "size", "color"],
            "line": ["x", "y", "color", "line_group", "facet_col"],
            "line_3d": ["x", "y", "z", "color"],
            "box": ["x", "y", "color", "facet_col"],
            "violin": ["x", "y", "color", "facet_col"],
            "scatter_matrix": ["dimensions", "size", "color"],
            "parallel_coordinates": ["dimensions"],
            "parallel_categories": ["dimensions"],
            "density_contour": ["x", "y", "color", "facet_col"],
        }

    @property
    def plot_args(self) -> dict:
        """A dict of possible plot options and their default values"""
        return OrderedDict(
            {
                "x": {
                    "options": self.columns,
                    "value": self.plot_options.get("x", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "y": {
                    "options": self.columns,
                    "value": self.plot_options.get("y", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "z": {
                    "options": self.columns,
                    "value": self.plot_options.get("z", self.columns[0]),
                    "multi": False,
                    "clearable": False,
                },
                "values": {
                    "options": self.numeric_columns,
                    "value": self.plot_options.get("values", self.numeric_columns[0]),
                    "multi": False,
                    "clearable": True,
                },
                "names": {
                    "options": self.columns,
                    "value": self.plot_options.get("names", None),
                    "multi": False,
                    "clearable": True,
                },
                "size": {
                    "options": self.numeric_columns,
                    "value": self.plot_options.get("size", None),
                    "multi": False,
                    "clearable": True,
                },
                "color": {
                    "options": self.columns,
                    "value": self.plot_options.get("color", None),
                    "multi": False,
                    "clearable": True,
                },
                "facet_col": {
                    "options": self.columns,
                    "value": self.plot_options.get("facet_col", None),
                    "multi": False,
                    "clearable": True,
                },
                "line_group": {
                    "options": self.columns,
                    "value": self.plot_options.get("line_group", None),
                    "multi": False,
                    "clearable": True,
                },
                "barmode": {
                    "options": ["stack", "group", "overlay", "relative"],
                    "value": self.plot_options.get("barmode", "stack"),
                    "multi": False,
                    "clearable": True,
                },
                "barnorm": {
                    "options": ["fraction", "percent"],
                    "value": self.plot_options.get("barnorm", None),
                    "multi": False,
                    "clearable": True,
                },
                "histnorm": {
                    "options": [
                        "percent",
                        "propability",
                        "density",
                        "propability density",
                    ],
                    "value": self.plot_options.get("histnorm", None),
                    "multi": False,
                    "clearable": True,
                },
                "trendline": {
                    "options": self.numeric_columns,
                    "value": None,
                    "multi": False,
                    "clearable": True,
                },
                "dimensions": {
                    "options": self.columns,
                    "value": self.plot_options.get("dimensions", self.columns),
                    "multi": True,
                    "clearable": True,
                },
            }
        )

    def filter_layout(self) -> Optional[list]:
        """Makes dropdowns for each dataframe column used for filtering."""
        if not self.use_filter:
            return None
        df = self.data
        dropdowns = []
        for col in self.filter_cols:
            if df[col].dtype in [np.float64, np.int64]:
                min_val = df[col].min()
                max_val = df[col].max()
                mean_val = df[col].mean()
                dropdowns.append(
                    wcc.Selectors(
                        label=col.lower().capitalize(),
                        children=dcc.RangeSlider(
                            id=self.uuid(f"filter-{col}"),
                            min=min_val,
                            max=max_val,
                            step=(max_val - min_val) / 10,
                            marks={
                                min_val: f"{min_val:.2f}",
                                mean_val: f"{mean_val:.2f}",
                                max_val: f"{max_val:.2f}",
                            },
                            value=[min_val, max_val],
                        ),
                    )
                )
            else:
                elements = list(self.data[col].unique())
                dropdowns.append(
                    wcc.Selectors(
                        label=col.lower().capitalize(),
                        children=wcc.Select(
                            id=self.uuid(f"filter-{col}"),
                            options=[{"label": i, "value": i} for i in elements],
                            value=elements
                            if self.filter_defaults is None
                            else [
                                element
                                for element in self.filter_defaults.get(col, elements)
                                if element in elements
                            ],
                            size=min(15, len(elements)),
                        ),
                    )
                )
        return wcc.Selectors(
            label="Filter options", open_details=False, children=dropdowns
        )

    def plot_option_layout(self) -> List[html.Div]:
        """Renders a dropdown widget for each plot option"""

        # The plot type dropdown is handled separate
        children = [
            wcc.Dropdown(
                label="Plot type",
                id=self.uuid("plottype"),
                clearable=False,
                options=[{"label": i, "value": i} for i in self.plots],
                value=self.plot_options.get("type", "scatter"),
            ),
        ]

        # Looping through all available plot options
        # and renders a dropdown widget
        for key, arg in self.plot_args.items():
            children.append(
                html.Div(
                    style=self.style_options_div_hidden,
                    id=self.uuid(f"div-{key}"),
                    children=[
                        wcc.Dropdown(
                            label=key,
                            id=self.uuid(f"dropdown-{key}"),
                            clearable=arg["clearable"],
                            options=[{"label": i, "value": i} for i in arg["options"]],
                            value=arg["value"],
                            multi=arg["multi"],
                        ),
                    ],
                )
            )
        return wcc.Selectors(
            label="Data and visualisation options",
            style=self.style_options_div,
            children=children,
        )

    @property
    def style_options_div(self) -> Dict[str, str]:
        """Style for active plot options"""
        return {"display": "grid"}

    @property
    def style_options_div_hidden(self) -> Dict[str, str]:
        """Style for hidden plot options"""
        return {"display": "none"}

    def axis_layout(self) -> html.Div:
        return wcc.Selectors(
            label="Plot axis options",
            open_details=False,
            children=[
                wcc.Label(
                    style={"fontSize": "0.9rem", "display": "block"},
                    children="Both min and max must be set to take effect",
                ),
                wcc.Label("X-axis", style={"fontWeight": "bold"}),
                html.Div(
                    style={"display": "flex"},
                    children=[
                        wcc.Label(style={"marginRight": "10px"}, children="Min: "),
                        dcc.Input(
                            style={"maxWidth": "80px"},
                            id=self.uuid("xaxis-min"),
                            type="number",
                            minLength=1,
                            placeholder="From data",
                        ),
                        wcc.Label(
                            style={
                                "marginLeft": "10px",
                                "marginRight": "10px",
                            },
                            children="Max: ",
                        ),
                        dcc.Input(
                            style={"maxWidth": "80px"},
                            id=self.uuid("xaxis-max"),
                            type="number",
                            minLength=1,
                            placeholder="From data",
                        ),
                    ],
                ),
                wcc.Label("Y-axis", style={"fontWeight": "bold"}),
                html.Div(
                    style={"display": "flex"},
                    children=[
                        wcc.Label(style={"marginRight": "10px"}, children="Min: "),
                        dcc.Input(
                            style={"maxWidth": "80px"},
                            id=self.uuid("yaxis-min"),
                            type="number",
                            minLength=1,
                            placeholder="From data",
                        ),
                        wcc.Label(
                            style={
                                "marginLeft": "10px",
                                "marginRight": "10px",
                            },
                            children="Max: ",
                        ),
                        dcc.Input(
                            style={"maxWidth": "80px"},
                            id=self.uuid("yaxis-max"),
                            type="number",
                            minLength=1,
                            placeholder="From data",
                        ),
                    ],
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        plot_options_style: dict = {"display": "none"} if self.lock else {}
        return html.Div(
            children=[
                wcc.FlexBox(
                    children=[
                        html.Div(
                            id=self.uuid("selector-row"),
                            children=[
                                html.Div(
                                    style=plot_options_style,
                                    children=self.plot_option_layout(),
                                ),
                                self.filter_layout(),
                                html.Div(
                                    style=plot_options_style,
                                    children=self.axis_layout(),
                                ),
                            ],
                        ),
                        wcc.Graph(
                            id=self.uuid("graph-id"),
                            style={"height": "80vh", "width": "75%"},
                        ),
                    ],
                )
            ]
        )

    @property
    def plot_output_callbacks(self) -> List[Output]:
        """Creates list of output dependencies for callback
        The outputs are the graph, and the style of the plot options"""
        outputs = []
        outputs.append(Output(self.uuid("graph-id"), "figure"))
        for plot_arg in self.plot_args.keys():
            outputs.append(Output(self.uuid(f"div-{plot_arg}"), "style"))
        return outputs

    @property
    def plot_input_callbacks(self) -> List[Input]:
        """Creates list of input dependencies for callback
        The inputs are the plot type and the current value
        for each plot option
        """
        inputs = []
        inputs.append(Input(self.uuid("plottype"), "value"))
        for plot_arg in self.plot_args.keys():
            inputs.append(Input(self.uuid(f"dropdown-{plot_arg}"), "value"))
        for filtcol in self.filter_cols:
            inputs.append(Input(self.uuid(f"filter-{filtcol}"), "value"))
        return inputs

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(self.plugin_data_output, self.plugin_data_requested)
        def _user_download_data(data_requested: Optional[int]) -> Optional[EncodedFile]:
            return (
                {
                    "filename": "table-plotter.csv",
                    "content": base64.b64encode(
                        get_data(self.csv_file).to_csv().encode()
                    ).decode("ascii"),
                    "mime_type": "text/csv",
                }
                if data_requested
                else None
            )

        @app.callback(
            self.plot_output_callbacks,
            Input(self.uuid("xaxis-min"), "value"),
            Input(self.uuid("xaxis-max"), "value"),
            Input(self.uuid("yaxis-min"), "value"),
            Input(self.uuid("yaxis-max"), "value"),
            self.plot_input_callbacks,
        )
        def _update_output(
            xaxis_min: Optional[float],
            xaxis_max: Optional[float],
            yaxis_min: Optional[float],
            yaxis_max: Optional[float],
            *args: Any,
        ) -> tuple:
            """Updates the graph and shows/hides plot options"""
            plot_type = args[0]
            # pylint: disable=protected-access
            plotfunc = getattr(px._chart_types, plot_type)
            plotargs = {}
            div_style = []
            data = self.data
            # Filter dataframe if filter columns are available
            if self.use_filter:
                plot_inputs = args[1 : -len(self.filter_cols)]
                filter_inputs = args[-len(self.filter_cols) :]
                data = filter_dataframe(data, self.filter_cols, filter_inputs)
            else:
                plot_inputs = args[1:]
            for name, plot_arg in zip(self.plot_args.keys(), plot_inputs):
                if plot_type in ["parallel_coordinates"] and name == "dimensions":
                    # This plot type only accepts numerical data
                    plot_arg = [val for val in plot_arg if val in self.numeric_columns]
                if name in self.plots[plot_type]:
                    plotargs[name] = plot_arg
                    div_style.append(self.style_options_div)

                    if (
                        name == "color"
                        and self.column_color_discrete_maps is not None
                        and plot_arg in self.column_color_discrete_maps
                        and "color_discrete_map"
                        in inspect.signature(plotfunc).parameters
                    ):
                        plotargs[
                            "color_discrete_map"
                        ] = self.column_color_discrete_maps.get(plot_arg)
                else:
                    div_style.append(self.style_options_div_hidden)
            figure: Figure = plotfunc(data, template=self.plotly_theme, **plotargs)
            if xaxis_min is not None and xaxis_max is not None:
                figure.update_layout(xaxis_range=[xaxis_min, xaxis_max])
            if yaxis_min is not None and yaxis_max is not None:
                figure.update_layout(yaxis_range=[yaxis_min, yaxis_max])

            return (figure, *div_style)


@CACHE.memoize()
@webvizstore
def get_data(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)


@CACHE.memoize()
def filter_dataframe(
    dframe: pd.DataFrame, columns: list, column_values: List[list]
) -> pd.DataFrame:
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            if df[col].dtype in [np.float64, np.int64]:
                df = df.loc[df[col].between(filt[0], filt[1])]
            else:
                df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    return df



from pathlib import Path

from dash import html

from .. import WebvizPluginABC
from ..webviz_assets import WEBVIZ_ASSETS


class ExampleAssets(WebvizPluginABC):
    def __init__(self, picture_path: Path, css_path: Path = None, js_path: Path = None):
        super().__init__()

        self.asset_url = WEBVIZ_ASSETS.add(picture_path)

        if css_path is not None:
            WEBVIZ_ASSETS.add(css_path)

        if js_path is not None:
            WEBVIZ_ASSETS.add(js_path)

    @property
    def layout(self) -> html.Img:
        return html.Img(src=self.asset_url)



from typing import Any, Dict, List, Optional, TypedDict, Union
from enum import Enum
from webviz_config.utils import callback_typecheck, ConversionError


def test_callback_typecheck() -> None:

    # pylint: disable=too-many-locals, too-many-statements
    class MyEnum(str, Enum):
        VALUE_1 = "value-1"

    class MyTypedDict(TypedDict):
        name: str
        year: int

    class DeepTypedDict(TypedDict):
        value: MyTypedDict

    ############################################################

    def expect_none(arg: None) -> None:
        assert arg is None

    callback_typecheck(expect_none)(None)

    ############################################################

    def expect_enum(arg: MyEnum) -> None:
        assert isinstance(arg, MyEnum)

    callback_typecheck(expect_enum)("value-1")

    ############################################################

    def expect_typed_dict(arg: MyTypedDict) -> None:
        types = [type(value) for value in arg.values()]
        assert (
            isinstance(arg, dict)
            and set(arg.keys())
            == set(MyTypedDict.__annotations__.keys())  # pylint: disable=no-member
            and set(types)
            == set(MyTypedDict.__annotations__.values())  # pylint: disable=no-member
        )

    callback_typecheck(expect_typed_dict)({"name": "Name", "year": 1990})

    # If any invalid key is given to a `TypedDict`, assert that a `ConversionError` is raised
    try:
        callback_typecheck(expect_typed_dict)({"name": "Name", "year2": 1990})
        assert False
    except ConversionError:
        pass
    except Exception:  # pylint: disable=broad-except
        assert False

    ############################################################

    def expect_deep_typed_dict(arg: DeepTypedDict) -> None:
        types = [type(value) for value in arg.values()]
        assert (
            isinstance(arg, dict)
            and set(arg.keys())
            == set(DeepTypedDict.__annotations__.keys())  # pylint: disable=no-member
            # NOTE: A `TypedDict` is a `dict` at runtime
            and set(types) == set([dict])
            and set(arg["value"].keys())
            == set(MyTypedDict.__annotations__.keys())  # pylint: disable=no-member
        )

    callback_typecheck(expect_deep_typed_dict)(
        {"value": {"name": "Name", "year": 1990}}
    )

    ############################################################

    def expect_int(arg: int) -> None:
        assert isinstance(arg, int)

    callback_typecheck(expect_int)("1")

    ############################################################

    def expect_float(arg: float) -> None:
        assert isinstance(arg, float)

    callback_typecheck(expect_float)("1.5")

    ############################################################

    def expect_str(arg: str) -> None:
        assert isinstance(arg, str)

    callback_typecheck(expect_str)(1)

    ############################################################

    def expect_list(arg: List[str]) -> None:
        assert isinstance(arg, list)
        assert isinstance(arg[0], str)

    callback_typecheck(expect_list)([1, 2, 3])

    ############################################################

    def expect_dict(arg: Dict[str, int]) -> None:
        assert isinstance(arg, dict)
        assert isinstance(list(arg.values())[0], int)
        assert isinstance(list(arg.keys())[0], str)

    callback_typecheck(expect_dict)({"1": 1})

    ############################################################

    def expect_optional(arg: Optional[str]) -> Optional[str]:
        return arg

    assert callback_typecheck(expect_optional)(None) is None
    assert isinstance(callback_typecheck(expect_optional)("string"), str)

    ############################################################

    def expect_union(arg: Union[str, int]) -> Union[str, int]:
        return arg

    assert isinstance(callback_typecheck(expect_union)("1"), str)
    assert isinstance(callback_typecheck(expect_union)(1), int)
    assert isinstance(callback_typecheck(expect_union)(1.5), str)

    ############################################################

    def expect_union_list(arg: Union[List[str], str]) -> Union[List[str], str]:
        return arg

    assert isinstance(callback_typecheck(expect_union_list)(["1", "2"]), list)
    assert isinstance(callback_typecheck(expect_union_list)(["1", "2"])[0], str)
    assert isinstance(callback_typecheck(expect_union_list)("1"), str)

    ############################################################

    def expect_optional_enum(arg: Optional[MyEnum]) -> Optional[MyEnum]:
        return arg

    assert callback_typecheck(expect_optional_enum)(None) is None
    assert isinstance(callback_typecheck(expect_optional_enum)("value-1"), MyEnum)

    ############################################################

    def expect_optional_string(arg: str) -> str:
        return arg

    try:
        callback_typecheck(expect_optional_string)(None)
        assert False
    except ConversionError:
        pass
    except Exception:  # pylint: disable=broad-except
        assert False

    ############################################################

    def expect_union_string_list(
        arg: Union[str, List[str], Dict[str, str], Optional[int]]
    ) -> Union[str, List[str], Dict[str, str], Optional[int]]:
        return arg

    assert callback_typecheck(expect_union_string_list)(["1", "2"]) == ["1", "2"]
    assert callback_typecheck(expect_union_string_list)("1") == "1"
    assert callback_typecheck(expect_union_string_list)({"1": "1"}) == {"1": "1"}
    assert callback_typecheck(expect_union_string_list)(None) is None
    assert callback_typecheck(expect_union_string_list)(1) == 1

    ############################################################

    def expect_optional_dict(arg: Optional[Dict[str, str]]) -> Optional[Dict[str, str]]:
        return arg

    assert callback_typecheck(expect_optional_dict)({"1": "1"}) == {"1": "1"}
    assert callback_typecheck(expect_optional_dict)(None) is None

    ############################################################

    def expect_optional_dict_without_types(arg: Optional[Dict]) -> Optional[Dict]:
        return arg

    assert callback_typecheck(expect_optional_dict_without_types)({"1": "1"}) == {
        "1": "1"
    }
    assert callback_typecheck(expect_optional_dict_without_types)(None) is None

    ############################################################

    def expect_optional_dict_with_any(arg: Optional[Dict[Any, Any]]) -> Optional[dict]:
        return arg

    assert callback_typecheck(expect_optional_dict_with_any)({"1": "1"}) == {"1": "1"}
    assert callback_typecheck(expect_optional_dict_with_any)(None) is None



from typing import Any

import pathlib

from dash.testing.composite import Browser
import dash
import webviz_core_components as wcc

from webviz_config.common_cache import CACHE
from webviz_config.themes import default_theme
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO, WebvizRunMode
from webviz_config import WebvizPluginABC

from ._webviz_ids import WebvizIds


class WebvizComposite(Browser):
    def __init__(self, server: Any, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.app = dash.Dash(__name__)
        self.server = server
        self.plugin: WebvizPluginABC
        self.init_app()

    def init_app(self) -> None:
        WEBVIZ_INSTANCE_INFO.initialize(
            dash_app=self.app,
            run_mode=WebvizRunMode.NON_PORTABLE,
            theme=default_theme,
            storage_folder=pathlib.Path(__file__).resolve().parent,
        )
        try:
            WEBVIZ_FACTORY_REGISTRY.initialize(None)
        except RuntimeError:
            pass

        self.app.css.config.serve_locally = True
        self.app.scripts.config.serve_locally = True
        self.app.config.suppress_callback_exceptions = True
        CACHE.init_app(self.app.server)

    def start_server(self, plugin: WebvizPluginABC, **kwargs: Any) -> None:
        """Start the local server with app."""

        self.app.layout = dash.html.Div(
            className=WebvizIds.LAYOUT_WRAPPER,
            children=[
                wcc.WebvizContentManager(
                    id=WebvizIds.CONTENT_MANAGER,
                    children=[
                        wcc.WebvizSettingsDrawer(
                            id=WebvizIds.SETTINGS_DRAWER,
                            children=plugin.get_all_settings(),
                        ),
                        wcc.WebvizPluginsWrapper(
                            id=WebvizIds.PLUGINS_WRAPPER,
                            children=plugin.plugin_layout(),
                        ),
                    ],
                ),
            ],
        )
        self.plugin = plugin
        # start server with app and pass Dash arguments
        self.server(self.app, **kwargs)

        # set the default server_url, it implicitly call wait_for_page
        self.server_url = self.server.url

    def toggle_webviz_settings_drawer(self) -> None:
        """Open the plugin settings drawer"""
        self.wait_for_element(WebvizIds.SETTINGS_DRAWER_TOGGLE_OPEN).click()

    def toggle_webviz_settings_group(self, settings_group_id: str) -> None:
        """Open the respective settings group in the settings drawer"""
        self.wait_for_element(
            f"#{settings_group_id} > .WebvizSettingsGroup__Title"
        ).click()

    def shared_settings_group_unique_component_id(
        self, settings_group_id: str, component_unique_id: str
    ) -> str:
        """Returns the element id of a component in a shared settings group"""
        unique_id = (
            self.plugin.shared_settings_group(settings_group_id)
            .component_unique_id(component_unique_id)
            .to_string()
        )
        return f"#{unique_id}"

    def view_settings_group_unique_component_id(
        self, view_id: str, settings_group_id: str, component_unique_id: str
    ) -> str:
        """Returns the element id of a component in a view settings group"""
        unique_id = (
            self.plugin.view(view_id)
            .settings_group(settings_group_id)
            .component_unique_id(component_unique_id)
            .to_string()
        )
        return f"#{unique_id}"



from typing import List

from dash.development.base_component import Component
from dash import html

import webviz_core_components as wcc

from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._views._plot import Kindness


class SharedSettingsGroup(SettingsGroupABC):
    class Ids(StrEnum):
        KINDNESS_SELECTOR = "kindness-selector"
        POWER_SELECTOR = "power-selector"

    def __init__(self) -> None:
        super().__init__("Shared settings")

    def layout(self) -> List[Component]:
        return [
            html.Div(
                children=[
                    wcc.Label("Kindness"),
                    wcc.RadioItems(
                        id=self.register_component_unique_id(
                            SharedSettingsGroup.Ids.KINDNESS_SELECTOR
                        ),
                        options=[
                            {
                                "label": a.value,
                                "value": a.value,
                            }
                            for a in Kindness
                        ],
                        value="friendly",
                    ),
                    wcc.Label("Power"),
                    wcc.RadioItems(
                        id=self.register_component_unique_id(
                            SharedSettingsGroup.Ids.POWER_SELECTOR
                        ),
                        options=[
                            {
                                "label": "2",
                                "value": 2,
                            },
                            {
                                "label": "3",
                                "value": 3,
                            },
                        ],
                        value=2,
                    ),
                ]
            )
        ]



from enum import Enum


class StrEnum(str, Enum):
    def __repr__(self) -> str:
        return self._value_  # pylint: disable=no-member

    def __str__(self) -> str:
        return self._value_  # pylint: disable=no-member



import warnings

import pytest

from webviz_config._dockerize._create_docker_setup import get_python_requirements


@pytest.mark.parametrize(
    "distributions, requirements",
    [
        (
            {
                "webviz-config": {
                    "download_url": "https://pypi.org/project/webviz-config",
                    "source_url": None,
                    "dist_version": "0.3.0",
                }
            },
            ["webviz-config==0.3.0"],
        ),
        (
            {
                "webviz-config": {
                    "download_url": None,
                    "source_url": None,
                    "dist_version": "0.3.0",
                }
            },
            [],
        ),
        (
            {
                "webviz-config": {
                    "download_url": None,
                    "source_url": "https://github.com/equinor/webviz-config",
                    "dist_version": "0.3.0",
                }
            },
            ["git+https://github.com/equinor/webviz-config@0.3.0"],
        ),
    ],
)
def test_derived_requirements(distributions, requirements):
    with warnings.catch_warnings(record=True) as record:
        warnings.simplefilter("always")
        assert set(requirements).issubset(get_python_requirements(distributions))
        assert len(record) == len(
            [
                metadata
                for metadata in distributions.values()
                if metadata["download_url"] is None and metadata["source_url"] is None
            ]
        )



import functools
from typing import Dict

from . import interactive_terminal
from . import radix_cli


def radix_configuration() -> Dict[str, str]:

    interactive_terminal.terminal_title("Radix cluster")
    radix_context = interactive_terminal.user_input_from_list(
        "WEBVIZ_RADIX_CONTEXT", "Radix context", ["production", "playground"]
    )
    radix_subdomain = "radix.equinor.com"
    if radix_context == "playground":
        radix_subdomain = "playground." + radix_subdomain

    interactive_terminal.terminal_title("Radix AD group")
    radix_ad_group = interactive_terminal.user_input_from_stdin(
        "WEBVIZ_RADIX_AD_GROUP",
        "AD group object ID",
    )

    interactive_terminal.terminal_title("Radix application name")
    radix_application_name = interactive_terminal.user_input_optional_reuse(
        "WEBVIZ_RADIX_APPLICATION",
        "Radix application",
        functools.partial(radix_cli.application_exists, context=radix_context),
        reuse_allowed=False,
    )

    return {
        "context": radix_context,
        "application_name": radix_application_name,
        "app_url": f"https://{radix_application_name}.app.{radix_subdomain}",
        "webhook_receiver_url": f"https://webhook.{radix_subdomain}/events/github",
        "ad_group": radix_ad_group,
    }



from pathlib import Path

from dash import html

from .. import WebvizPluginABC
from ..webviz_assets import WEBVIZ_ASSETS


class EmbedPdf(WebvizPluginABC):
    """Embeds a given PDF file into the page.

!> Webviz does not scan your PDF for malicious code. Make sure it comes from a trusted source.
---

* **`pdf_file`:** Path to the PDF file to include. Either absolute path or \
  relative to the configuration file.
* **`height`:** Height of the PDF object (in percent of viewport height).
* **`width`:** Width of the PDF object (in percent of available space).

"""

    def __init__(self, pdf_file: Path, height: int = 80, width: int = 100):

        super().__init__()

        self.pdf_url = WEBVIZ_ASSETS.add(pdf_file)
        self.height = height
        self.width = width

    @property
    def layout(self) -> html.Embed:

        style = {"height": f"{self.height}vh", "width": f"{self.width}%"}

        return html.Embed(src=self.pdf_url, style=style, type="application/pdf")



from typing import Any, Generator

# pylint: disable=too-few-public-methods
class MissingWebvizTesting:
    def __init__(self, **kwargs: Any) -> None:
        raise RuntimeError(
            "webviz_config[tests] was not installed. "
            "Please install to use the webviz testing fixtures."
        )


try:
    from ._composite import WebvizComposite

except ImportError:
    WebvizComposite = MissingWebvizTesting  # type: ignore



import sys
import subprocess  # nosec


def test_portable(dash_duo, tmp_path):
    # Build a portable webviz from config file
    appdir = tmp_path / "app"
    subprocess.call(  # nosec
        ["webviz", "build", "basic_example.yaml", "--portable", appdir], cwd="examples"
    )

    # Import generated app
    sys.path.append(str(appdir))
    from webviz_app import app  # pylint: disable=import-error, import-outside-toplevel

    # Start and test app
    dash_duo.start_server(app)
    for page in [
        "markdown-example",
        "table-example",
        "pdf-example",
        "syntax-highlighting-example",
        "plot-a-table",
        "pivot-table",
    ]:
        dash_duo.wait_for_element(f".Menu__Page[href='/{page}']").click()
    assert not dash_duo.get_logs(), "browser console should contain no error"



from typing import Any, Generator

import pytest

from webviz_config.testing import WebvizComposite


@pytest.fixture
def _webviz_duo(request: Any, dash_thread_server: Any, tmpdir: Any) -> Generator:
    with WebvizComposite(
        dash_thread_server,
        browser=request.config.getoption("webdriver"),
        remote=request.config.getoption("remote"),
        remote_url=request.config.getoption("remote_url"),
        headless=request.config.getoption("headless"),
        options=request.config.hook.pytest_setup_options(),
        download_path=tmpdir.mkdir("download").strpath,
        percy_assets_root=request.config.getoption("percy_assets"),
        percy_finalize=request.config.getoption("nopercyfinalize"),
        pause=request.config.getoption("pause"),
    ) as duo:
        yield duo



from webviz_config.testing import WebvizComposite
from webviz_config.generic_plugins._example_wlf_plugin import ExampleWlfPlugin


def test_example_wlf_plugin(
    _webviz_duo: WebvizComposite,
) -> None:
    plugin = ExampleWlfPlugin(title="hello")

    _webviz_duo.start_server(plugin)

    _webviz_duo.toggle_webviz_settings_drawer()

    _webviz_duo.toggle_webviz_settings_group(
        plugin.view("plot-view").settings_group_unique_id("plot-settings")
    )

    component_id = _webviz_duo.view_settings_group_unique_component_id(
        view_id="plot-view",
        settings_group_id="plot-settings",
        component_unique_id="coordinates-selector",
    )

    _webviz_duo.wait_for_contains_text(component_id, "x - y")
    assert not _webviz_duo.get_logs()



from enum import Enum


class WebvizIds(str, Enum):
    LAYOUT_WRAPPER = "layoutWrapper"
    CONTENT_MANAGER = "webviz-content-manager"
    SETTINGS_DRAWER = "settings-drawer"
    PLUGINS_WRAPPER = "plugins-wrapper"
    SETTINGS_DRAWER_TOGGLE_OPEN = ".WebvizSettingsDrawer__ToggleOpen"



import sys
from importlib.metadata import EntryPoint, entry_points

from .. import WebvizConfigTheme
from ._default_theme import default_theme

installed_themes = {default_theme.theme_name: default_theme}

__all__ = ["installed_themes"]


def process_entry_point(entry_point: EntryPoint):
    theme = entry_point.load()

    globals()[entry_point.name] = theme
    __all__.append(entry_point.name)

    if isinstance(theme, WebvizConfigTheme):
        installed_themes[theme.theme_name] = theme


if sys.version_info < (3, 10, 0):
    eps = entry_points().get("webviz_config_themes", [])
else:
    eps = entry_points().select(group="webviz_config_themes")

for ep in eps:
    process_entry_point(ep)



from pathlib import Path
from typing import cast

import pytest
from dash import Dash

from webviz_config import WebvizConfigTheme
from webviz_config.webviz_instance_info import WebvizInstanceInfo, WebvizRunMode


def test_construction_and_basic_access() -> None:
    my_dash_app = Dash("dummyAppName")
    my_run_mode = WebvizRunMode.NON_PORTABLE
    my_theme = WebvizConfigTheme("dummyThemeName")
    my_storage_folder = Path("dummyPath")

    instance_info = WebvizInstanceInfo()
    instance_info.initialize(
        dash_app=my_dash_app,
        run_mode=my_run_mode,
        theme=my_theme,
        storage_folder=my_storage_folder,
    )

    assert instance_info.dash_app is my_dash_app
    assert instance_info.run_mode is WebvizRunMode.NON_PORTABLE
    assert instance_info.theme.__dict__ == my_theme.__dict__
    assert instance_info.storage_folder == Path("dummyPath")


def test_that_construction_with_invalid_types_throw() -> None:
    dash_app = Dash("dummyAppName")
    run_mode = WebvizRunMode.NON_PORTABLE
    theme = WebvizConfigTheme("dummyThemeName")
    storage_folder = Path("dummyPath")

    with pytest.raises(TypeError):
        WebvizInstanceInfo().initialize(
            dash_app=cast(Dash, None),
            run_mode=run_mode,
            theme=theme,
            storage_folder=storage_folder,
        )
    with pytest.raises(TypeError):
        WebvizInstanceInfo().initialize(
            dash_app=dash_app,
            run_mode=cast(WebvizRunMode, None),
            theme=theme,
            storage_folder=storage_folder,
        )
    with pytest.raises(TypeError):
        WebvizInstanceInfo().initialize(
            dash_app=dash_app,
            run_mode=run_mode,
            theme=cast(WebvizConfigTheme, None),
            storage_folder=storage_folder,
        )
    with pytest.raises(TypeError):
        WebvizInstanceInfo().initialize(
            dash_app=dash_app,
            run_mode=run_mode,
            theme=theme,
            storage_folder=cast(Path, None),
        )


def test_immutability() -> None:
    my_dash_app = Dash("dummyAppName")
    my_run_mode = WebvizRunMode.NON_PORTABLE
    my_theme = WebvizConfigTheme("dummyThemeName")
    my_storage_folder = Path("dummyPath")

    instance_info = WebvizInstanceInfo()
    instance_info.initialize(
        dash_app=my_dash_app,
        run_mode=my_run_mode,
        theme=my_theme,
        storage_folder=my_storage_folder,
    )

    # Ony allowed to initialize once
    with pytest.raises(RuntimeError):
        instance_info.initialize(
            dash_app=my_dash_app,
            run_mode=my_run_mode,
            theme=my_theme,
            storage_folder=my_storage_folder,
        )

    # This is ok and necessary since we want to share the actual dash app
    assert instance_info.dash_app is my_dash_app

    # This two are also ok since integer enums and Path themselves is immutable
    assert instance_info.run_mode is my_run_mode
    assert instance_info.storage_folder is my_storage_folder

    returned_theme = instance_info.theme
    assert returned_theme is not my_theme
    assert isinstance(returned_theme, WebvizConfigTheme)
    assert returned_theme.__dict__ == my_theme.__dict__
    my_theme.theme_name = "MODIFIED"
    assert returned_theme.__dict__ != my_theme.__dict__

    with pytest.raises(AttributeError):
        # pylint: disable=assigning-non-slot
        instance_info.some_new_attribute = "myAttributeValue"  # type: ignore[attr-defined]



import warnings

import flask_caching


class Cache(flask_caching.Cache):
    def __init__(
        self,
    ) -> None:

        super().__init__(config={"CACHE_TYPE": "simple", "DEFAULT_CACHE_TIMEOUT": 3600})

    @property
    def TIMEOUT(self) -> int:  # pylint: disable=invalid-name
        warnings.warn(
            "Default cache timeout is now initialized directly, and the TIMEOUT "
            "attribute is therefore deprecated, and will be removed. The timeout "
            "argument can still be used as input to @CACHE.memoize() with an input "
            "if you want to overwride the 3600 (seconds) webviz default.",
            DeprecationWarning,
        )
        return 3600


CACHE = Cache()



import time
from pathlib import Path

import dash
from dash.testing.composite import DashComposite

from webviz_config import WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.themes import default_theme
from webviz_config.generic_plugins import _table_plotter


def test_table_plotter(dash_duo: DashComposite) -> None:

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    webviz_settings = WebvizSettings({}, default_theme)
    csv_file = Path("./tests/data/example_data.csv")
    page = _table_plotter.TablePlotter(app, webviz_settings, csv_file)
    app.layout = page.layout
    dash_duo.start_server(app)

    # Wait for the app to render(there is probably a better way...)
    time.sleep(5)

    # Checking that no plot options are defined
    assert not page.plot_options
    # Check that filter is not active
    assert not page.use_filter

    # Checking that the correct plot type is initialized
    plot_dd = dash_duo.find_element("#" + page.uuid("plottype"))
    assert plot_dd.text == "scatter"

    # Checking that only the relevant options are shown
    for plot_option in page.plot_args.keys():
        plot_option_dd = dash_duo.find_element("#" + page.uuid(f"div-{plot_option}"))
        if plot_option not in page.plots["scatter"]:
            assert plot_option_dd.get_attribute("style") == "display: none;"

    # Checking that options are initialized correctly
    for option in ["x", "y"]:
        plot_option_dd = dash_duo.find_element("#" + page.uuid(f"dropdown-{option}"))
        assert plot_option_dd.text == "Well"


def test_table_plotter_filter(dash_duo: DashComposite) -> None:

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    webviz_settings = WebvizSettings({}, default_theme)
    csv_file = Path("./tests/data/example_data.csv")
    page = _table_plotter.TablePlotter(
        app, webviz_settings, csv_file, filter_cols=["Well"]
    )
    app.layout = page.layout
    dash_duo.start_server(app)

    # Wait for the app to render(there is probably a better way...)
    time.sleep(5)

    # Checking that no plot options are defined
    assert not page.plot_options
    # Check that filter is active
    assert page.use_filter
    assert page.filter_cols == ["Well"]

    # Checking that the correct plot type is initialized
    plot_dd = dash_duo.find_element("#" + page.uuid("plottype"))
    assert plot_dd.text == "scatter"

    # Checking that options are initialized correctly
    for option in ["x", "y"]:
        plot_option_dd = dash_duo.find_element("#" + page.uuid(f"dropdown-{option}"))
        assert plot_option_dd.text == "Well"


def test_initialized_table_plotter(dash_duo: DashComposite) -> None:

    app = dash.Dash(__name__)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    webviz_settings = WebvizSettings({}, default_theme)
    csv_file = Path("./tests/data/example_data.csv")
    plot_options = dict(
        x="Well",
        y="Initial reservoir pressure (bar)",
        size="Average permeability (D)",
        facet_col="Segment",
    )

    page = _table_plotter.TablePlotter(
        app, webviz_settings, csv_file, lock=True, plot_options=plot_options
    )
    app.layout = page.layout
    dash_duo.start_server(app)

    # Wait for the app to render(there is probably a better way...)

    # Checking that plot options are defined
    assert page.plot_options == plot_options
    assert page.lock



"""These are the basic Webviz configuration plugins, distributed through
the utility itself.
"""

import abc
from importlib.metadata import distributions

from ._utils import load_webviz_plugins_with_metadata, PluginProjectMetaData


(
    PLUGIN_METADATA,
    PLUGIN_PROJECT_METADATA,
    plugin_entrypoints,
) = load_webviz_plugins_with_metadata(distributions())

__all__ = list(plugin_entrypoints.keys())


def __getattr__(name: str) -> abc.ABC:
    """Lazy load plugins, i.e. only import/load when a given plugin is requested."""

    if name in __all__:
        return plugin_entrypoints[name].load()
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import os
import datetime
from typing import Tuple

import msal
import flask
from werkzeug.middleware.proxy_fix import ProxyFix


class Oauth2:
    """Oauth2 authorization Code grant flow"""

    def __init__(self, app: flask.app.Flask):
        self._app = app

        self.configure_proxy_trust()

        # Azure AD app registration info (currently the values are taken from environment variables)
        self._tenant_id = os.environ["WEBVIZ_TENANT_ID"]
        self._client_id = os.environ["WEBVIZ_CLIENT_ID"]
        self._client_secret = os.environ["WEBVIZ_CLIENT_SECRET"]
        self._scope = os.environ["WEBVIZ_SCOPE"]

        # Initiate msal
        self._msal_app = msal.ConfidentialClientApplication(
            client_id=self._client_id,
            client_credential=self._client_secret,
            authority=f"https://login.microsoftonline.com/{self._tenant_id}",
        )
        self._accounts = self._msal_app.get_accounts()

        # Initiate oauth2 endpoints
        self.set_oauth2_endpoints()

    def configure_proxy_trust(self) -> None:
        """Configure"""
        _proxy_settings = {
            "x_for": os.environ.get("WEBVIZ_X_FORWARDED_FOR"),
            "x_proto": os.environ.get("WEBVIZ_X_FORWARDED_PROTO"),
            "x_host": os.environ.get("WEBVIZ_X_FORWARDED_HOST"),
            "x_port": os.environ.get("WEBVIZ_X_FORWARDED_PORT"),
            "x_prefix": os.environ.get("WEBVIZ_X_FORWARDED_PREFIX"),
        }

        if any(x is not None for x in _proxy_settings.values()):
            proxy_settings = {k: int(v) if v else 0 for k, v in _proxy_settings.items()}
            self._app.wsgi_app = ProxyFix(self._app.wsgi_app, **proxy_settings)  # type: ignore

    def set_oauth2_endpoints(self) -> None:
        """/login and /auth-return endpoints are added for Oauth2 authorization
        code flow.

        At the end of the flow, a session cookie containing a valid access token
        and its expiration date is created. This flask session object can be
        accessed from Webviz plugin.

        To get the access token: flask.session.get("access_token")
        To get the expiration date: flask.session.get("expiration_date")

        An Azure AD application should be registered, and the following environment
        variables should be set: WEBVIZ_TENANT_ID, WEBVIZ_CLIENT_ID,
        WEBVIZ_CLIENT_SECRET, WEBVIZ_SCOPE.
        """

        @self._app.route("/login")
        def _login_controller():  # type: ignore[no-untyped-def]
            redirect_uri = get_auth_redirect_uri()

            # First leg of Oauth2 authorization code flow
            auth_url = self._msal_app.get_authorization_request_url(
                scopes=[self._scope], redirect_uri=redirect_uri
            )
            return flask.redirect(auth_url)

        @self._app.route("/auth-return")
        def _auth_return_controller():  # type: ignore[no-untyped-def]
            redirect_uri = get_auth_redirect_uri()
            returned_query_params = flask.request.args

            # There is an error from the first leg of Oauth2 authorization code flow
            if "error" in returned_query_params:
                error_description = returned_query_params.get("error_description")
                print("Error description:", error_description)
                redirect_error_uri = flask.url_for("_error_controller")
                return flask.redirect(redirect_error_uri)

            code = returned_query_params.get("code")

            # Second leg of Oauth2 authorization code flow
            tokens_result = self._msal_app.acquire_token_by_authorization_code(
                code=code, scopes=[self._scope], redirect_uri=redirect_uri
            )
            expires_in = tokens_result.get("expires_in")
            expiration_date = datetime.datetime.now(
                datetime.timezone.utc
            ) + datetime.timedelta(seconds=expires_in - 60)
            print("Access token expiration date (UTC):", expiration_date)

            # Set expiration date in the session
            flask.session["expiration_date"] = expiration_date

            # Set access token in the session
            flask.session["access_token"] = tokens_result.get("access_token")

            return flask.redirect(flask.request.url_root)

        @self._app.route("/error")
        def _error_controller():  # type: ignore[no-untyped-def]
            return "Error"

    def set_oauth2_before_request_decorator(self) -> None:
        """Check access token existence in session cookie before every request.
        If it does not exist, the browser is redirected to /login endpoint.

        If access token exists, its expiration date is checked in session cookie.
        If the current date exceeds its expiration date, a new access token is
        retrieved and set in the session cookie. A new expiration date is also
        set in the session cookie.
        """

        @self._app.before_request
        def _check_access_token():  # type: ignore[no-untyped-def]
            return self.check_access_token()

    # pylint: disable=inconsistent-return-statements
    def check_access_token(self):  # type: ignore[no-untyped-def]
        self.check_and_set_token_expiry()

        # If the client session does not contain access token, redirect to /login
        is_redirected, redirect_url = self.is_empty_token()
        if is_redirected:
            return flask.redirect(redirect_url)

    @staticmethod
    def is_empty_token() -> Tuple[bool, str]:
        if (
            not flask.session.get("access_token")
            and flask.request.endpoint != "_login_controller"
            and flask.request.endpoint != "_auth_return_controller"
        ):
            login_uri = get_login_uri()
            return True, login_uri

        return False, ""

    def check_and_set_token_expiry(self) -> None:
        if flask.session.get("access_token"):
            expiration_date = flask.session["expiration_date"]
            current_date = datetime.datetime.now(datetime.timezone.utc)
            if current_date > expiration_date:
                # Access token has expired
                print("Access token has expired, trying to refresh.")
                try:
                    raise Exception(
                        "Disable refreshing, it is broken and a security risk"
                    )
                    # pylint: disable=W0101
                    access_token, expiration_date = self.refresh_token_if_possible()
                    flask.session["access_token"] = access_token
                    flask.session["expiration_date"] = expiration_date
                except Exception as exc:  # pylint: disable=broad-except
                    # Catch all and any error here,
                    # since in all cases an error means we want to force a loud login
                    print(
                        f"Warning: error {exc} encountered when trying to refresh access token.\n"
                        "Clearing any stored access token info to force re-login"
                    )
                    if "access_token" in flask.session:
                        flask.session.pop("access_token")
                    if "expiration_date" in flask.session:
                        flask.session.pop("expiration_date")

    def refresh_token_if_possible(self) -> Tuple[str, datetime.datetime]:
        if not self._accounts:
            self._accounts = self._msal_app.get_accounts()
        renewed_tokens_result = self._msal_app.acquire_token_silent(
            scopes=[self._scope], account=self._accounts[0]
        )
        expires_in = renewed_tokens_result.get("expires_in")
        new_expiration_date = datetime.datetime.now(
            datetime.timezone.utc
        ) + datetime.timedelta(seconds=expires_in - 60)
        print("New access token expiration date (UTC):", new_expiration_date)

        return renewed_tokens_result.get("access_token"), new_expiration_date


def get_login_uri() -> str:
    return flask.url_for("_login_controller")


def get_auth_redirect_uri() -> str:
    return flask.url_for("_auth_return_controller", _external=True)



import warnings
from typing import Optional

import requests


def pip_git_url(
    version: str, source_url: str, git_pointer: Optional[str] = None
) -> str:
    """All webviz-config plugin projects are encouraged to be open source
    and on PyPI. Those that are not are encouraged to be on a git solution,
    and use quite standard setuptools_scm to give a runtime available version
    to the Python package.

    setuptool_scm enables us to get information regarding which hash/tag of the
    git repository to download + install in the Docker image, as installation from a
    clean repository state should give one of the following versions:

    no distance to tag:
        {tag}
    distance:
        {next_version}.dev{distance}+g{revision hash}

    If git_pointer argument is not None, this git ponter/reference is used instead
    of derived commit hash/tag from version.
    """

    source_url = source_url.rstrip("/")

    if git_pointer is None:
        git_pointer = version.split("+g")[1] if "+" in version else version

    if requests.get(source_url).status_code != 200:
        warnings.warn(
            f"Could not find {source_url}, assuming private repository. Changing to "
            "SSH URL. If building Docker image you will need to provide a deploy key."
        )
        for string in ["https://github.com/", "https://www.github.com/"]:
            source_url = source_url.replace(string, "ssh://git@github.com/")

    return f"git+{source_url}@{git_pointer}"



from pathlib import Path
from unittest import mock

import dash

from webviz_config.common_cache import CACHE
from webviz_config.generic_plugins import _syntax_highlighter


def test_syntax_highlighter(dash_duo):
    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    code_file = Path("./tests/data/basic_example.yaml")
    with mock.patch(
        "webviz_config.generic_plugins._syntax_highlighter.get_path"
    ) as mock_path:
        mock_path.return_value = code_file
        page = _syntax_highlighter.SyntaxHighlighter(app, code_file)
        app.layout = page.layout
        dash_duo.start_server(app)
        assert not dash_duo.get_logs(), "browser console should contain no error"



import copy
import pathlib
from typing import Any

from ._build_docs import get_plugin_documentation


JSON_SCHEMA = {
    "$id": "https://github.com/equinor/webviz-config",
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Webviz",
    "type": "object",
    "properties": {
        "title": {"description": "Title of your Webviz application.", "type": "string"},
        "shared_settings": {"type": "object"},
        "options": {
            "type": "object",
            "menu": {
                "description": "Define the menu options.",
                "type": "object",
                "properties": {
                    "show_logo": {
                        "description": "State if a logo shall be shown in the menu.",
                        "type": "boolean",
                    },
                    "bar_position": {
                        "description": "Define where the menu bar shall be positioned:"
                        " left, top, right, bottom.",
                        "type": "string",
                    },
                    "drawer_position": {
                        "description": "Define where the menu drawer shall be positioned:"
                        " left or right.",
                        "type": "string",
                    },
                    "initially_pinned": {
                        "description": "State if the menu shall be pinned when initially showing.",
                        "type": "boolean",
                    },
                    "initially_collapsed": {
                        "description": "State if all groups in menu shall initially be collapsed",
                        "type": "boolean",
                    },
                    "homepage": {
                        "description": """
                        Set a custom page as homepage to which the user returns when clicking on the logo.
                        Use the page's title.
                        """,
                        "type": "string",
                    },
                },
                "additionalProperties": False,
            },
            "plotly_theme": {
                "type": "object",
                "description": """
                Option to define modifications to the theme's Plotly figure layout.
                Examples are e.g. axis formatting and color palettes. Will be merged
                with existing theme layout, with options defined here prioritized if
                conflicting with the existing theme. The layout is defined as a
                dictionary, for details see:
                https://plotly.com/python/reference/layout/
                """,
            },
        },
        "layout": {
            "description": "Define the pages (and potential sections and groups)"
            " in your Webviz application.",
            "type": "array",
            "minLength": 1,
            "items": {
                "oneOf": [
                    {
                        "type": "object",
                        "properties": {
                            "section": {
                                "description": "Title of the section.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally set an icon for the section.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Define the pages (and potential subgroups)"
                                " of this section.",
                                "type": "array",
                                "minLength": 1,
                                "items": {
                                    "type": "object",
                                    "oneOf": [
                                        {"$ref": "#/properties/layout/items/oneOf/1"},
                                        {"$ref": "#/properties/layout/items/oneOf/2"},
                                    ],
                                },
                            },
                        },
                        "required": ["section", "content"],
                        "additionalProperties": False,
                    },
                    {
                        "type": "object",
                        "properties": {
                            "group": {
                                "description": "Title of the group.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally define an icon for the group.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Content of the group.",
                                "type": "array",
                                "minLength": 1,
                                "items": {
                                    "oneOf": [
                                        {"$ref": "#/properties/layout/items/oneOf/1"},
                                        {"$ref": "#/properties/layout/items/oneOf/2"},
                                    ]
                                },
                            },
                        },
                        "required": ["group", "content"],
                        "additionalProperties": False,
                    },
                    {
                        "type": "object",
                        "properties": {
                            "page": {
                                "description": "Title of the page.",
                                "type": "string",
                            },
                            "icon": {
                                "description": "Optionally define an icon for the page.",
                                "type": "string",
                            },
                            "content": {
                                "description": "Content of the page.",
                                "type": "array",
                                "minLength": 1,
                                "items": {"oneOf": [{"type": "string"}]},
                            },
                        },
                        "required": ["page", "content"],
                        "additionalProperties": False,
                    },
                ],
            },
        },
    },
    "required": ["title", "layout"],
    "additionalProperties": False,
}

CONTACT_PERSON = {
    "contact_person": {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "email": {"type": "string"},
            "phone": {"type": ["integer", "string"]},
        },
        "additionalProperties": False,
    }
}


def create_schema() -> dict:
    def json_type(typehint: Any) -> dict:
        # pylint: disable=too-many-return-statements
        if typehint == list:
            return {"type": "array", "items": {}}
        if typehint in (str, pathlib.Path):
            return {"type": "string"}
        if typehint == bool:
            return {"type": "boolean"}
        if typehint == int:
            return {"type": "integer"}
        if typehint == float:
            return {"type": "number"}
        if typehint == dict:
            return {"type": "object"}
        return {}

    json_schema = copy.deepcopy(JSON_SCHEMA)

    # fmt: off
    content_schemas = json_schema["properties"]["layout"][  # type: ignore
        "items"]["oneOf"][2]["properties"]["content"]["items"]["oneOf"]
    # fmt: on

    for package_doc in get_plugin_documentation().values():
        for plugin_doc in package_doc["plugins"]:
            content_schemas.append(
                {
                    "type": "object",
                    "properties": {
                        plugin_doc["name"]: {
                            "description": plugin_doc["description"]
                            or " PLUGIN MISSING DESCRIPTION ",
                            "type": "object",
                            "properties": {
                                **{
                                    arg_name: json_type(arg_info.get("typehint"))
                                    for arg_name, arg_info in plugin_doc[
                                        "arg_info"
                                    ].items()
                                },
                                **CONTACT_PERSON,
                            },
                            "required": [
                                arg_name
                                for arg_name, arg_info in plugin_doc["arg_info"].items()
                                if arg_info["required"]
                            ],
                            "additionalProperties": False,
                        }
                    },
                    "required": [plugin_doc["name"]],
                    "additionalProperties": False,
                    "minProperties": 1,
                    "maxProperties": 1,
                }
            )

    return json_schema
